@@ -128,16 +128,17 @@ properties persist through refontification."
     (with-temp-buffer
       (insert-buffer-substring-no-properties org-buffer start end)
       (insert " ")                      ; Add space to ensure property change
-      (if (symbolp path-or-mode)
-          (setq lang-mode path-or-mode)
-        (let ((buffer-file-name path-or-mode))
-          (setq lang-mode
-                (or (cdr (assoc-string
-                          (concat
-                           "\\." (file-name-extension path-or-mode) "\\'")
-                          auto-mode-alist))
-                    (progn (set-auto-mode t) major-mode)))))
-      (delay-mode-hooks (funcall lang-mode))
+      (delay-mode-hooks
+        (if (symbolp path-or-mode)
+            (setq lang-mode path-or-mode)
+          (let ((buffer-file-name path-or-mode))
+            (setq lang-mode
+                  (or (cdr (assoc-string
+                            (concat
+                             "\\." (file-name-extension path-or-mode) "\\'")
+                            auto-mode-alist))
+                      (progn (set-auto-mode t) major-mode)))))
+        (funcall lang-mode))
       (font-lock-ensure)
       (let ((pos (point-min)))
         (while (< pos (1- (point-max))) ; Skip the added space
@@ -220,6 +221,31 @@ ARG-VALUES is the list of arguments for the tool call."
      inner-from (1- (point)) 'font-lock-face (gptel-agent--block-bg))
     (gptel-agent--confirm-overlay from (point) t)))
 
+(defun gptel-agent--execute-bash (callback command)
+  "Execute COMMAND asynchronously in bash and call CALLBACK with output.
+
+CALLBACK is called with the command output string when the process finishes.
+COMMAND is the bash command string to execute."
+  (let* ((output-buffer (generate-new-buffer " *gptel-agent-bash*"))
+         (proc (make-process
+                :name "gptel-agent-bash"
+                :buffer output-buffer
+                :command (list "bash" "-c" command)
+                :connection-type 'pipe
+                :sentinel
+                (lambda (process _event)
+                  (when (memq (process-status process) '(exit signal))
+                    (let* ((exit-code (process-exit-status process))
+                           (output (with-current-buffer (process-buffer process)
+                                     (buffer-string))))
+                      (kill-buffer (process-buffer process))
+                      (funcall callback
+                               (if (zerop exit-code)
+                                   output
+                                 (format "Command failed with exit code %d:\nSTDOUT+STDERR:\n%s"
+                                         exit-code output)))))))))
+    proc))
+
 ;;; Web tools
 
 (defun gptel-agent--fetch-with-timeout (url url-cb tool-cb failed-msg &rest args)
@@ -930,30 +956,35 @@ CONTEXT-LINES specifies the number of lines of context to show
   around each match (0-15 inclusive, defaults to 0).
 
 Returns a string containing matches grouped by file, with line numbers
-and optional context. Results are sorted by modification time."
+and optional context.  Results are limited to 1000 or fewer matches per
+file.  Results are sorted by modification time."
   (unless (file-readable-p path)
     (error "Error: File or directory %s is not readable" path))
   (let ((grepper (or (executable-find "rg") (executable-find "grep"))))
     (unless grepper
       (error "Error: ripgrep/grep not available, this tool cannot be used"))
     (with-temp-buffer
-      (let* ((args
+      (let* ((cmd (file-name-sans-extension (file-name-nondirectory grepper)))
+             (args
               (cond
-               ((string-suffix-p "rg" grepper)
+               ((string= "rg" cmd)
                 (delq nil (list "--sort=modified"
                                 (and (natnump context-lines)
                                      (format "--context=%d" context-lines))
                                 (and glob (format "--glob=%s" glob))
-                                ;; "--files-with-matches" "--max-count=10"
+                                ;; "--files-with-matches"
+                                "--max-count=1000"
                                 "--heading" "--line-number" "-e" regex
                                 (expand-file-name (substitute-in-file-name path)))))
-               ((string-suffix-p "grep" grepper)
+               ((string= "grep" cmd)
                 (delq nil (list "--recursive"
                                 (and (natnump context-lines)
                                      (format "--context=%d" context-lines))
                                 (and glob (format "--include=%s" glob))
+                                "--max-count=1000"
                                 "--line-number" "--regexp" regex
-                                (expand-file-name (substitute-in-file-name path)))))))
+                                (expand-file-name (substitute-in-file-name path)))))
+               (t (error "Error: failed to identify grepper"))))
              (exit-code (apply #'call-process grepper nil '(t t) nil args)))
         (when (/= exit-code 0)
           (goto-char (point-min))
@@ -1217,16 +1248,7 @@ Error details: %S"
 
 (gptel-make-tool
  :name "Bash"
- :function (lambda (command)
-             "Execute a bash command and return its output.
-
-COMMAND is the bash command string to execute."
-             (with-temp-buffer
-               (let* ((exit-code (call-process "bash" nil (current-buffer) nil "-c" command))
-                      (output (buffer-string)))
-                 (if (zerop exit-code)
-                     output
-                   (format "Command failed with exit code %d:\nSTDOUT+STDERR:\n%s" exit-code output)))))
+ :function #'gptel-agent--execute-bash
  :description "Execute Bash commands.
 
 This tool provides access to a Bash shell with GNU coreutils (or equivalents) available.
@@ -1257,7 +1279,8 @@ Can include pipes and standard shell operators.
 Example: 'ls -la | head -20' or 'grep -i error app.log | tail -50'"))
  :category "gptel-agent"
  :confirm t
- :include t)
+ :include t
+ :async t)
 
 (gptel-make-tool
  :name "Eval"
@@ -1454,7 +1477,8 @@ specific location with no changes to the surrounding context."
 - -1 to insert at the end."
            :type integer)
          ( :name "new_str"
-           :description "String to insert at `line_number`."))
+           :description "String to insert at `line_number`."
+           :type string))
  :category "gptel-agent"
  :confirm t
  :include t)
@@ -1492,6 +1516,7 @@ Consider using the more granular tools \"Insert\" or \"Edit\" first."
 
 - Supports glob patterns like \"*.md\" or \"*test*.py\".
   The glob applies to the basename of the file (with extension).
+- Does not support double wildcard \"**/*\".
 - Returns matching file paths at all depths sorted by modification time.
   Limit the depth of the search by providing the `depth` argument.
 - When you are doing an open ended search that may require multiple rounds
